from .util_io import getFileHandle
from .. import settings as og_settings
from shapely import wkt
import os
import csv


def _outputMesoNodes(mesonet, mesonode_filepath, projection, encoding):
    outfile = getFileHandle(mesonode_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['node_id', 'zone_id', 'x_coord', 'y_coord', 'macro_node_id', 'macro_link_id', 'activity_type', 'is_boundary'])
    for _, mesonode in mesonet.node_dict.items():
        if projection:
            x_coord = round(mesonode.geometry_xy.x, og_settings.local_coord_precision)
            y_coord = round(mesonode.geometry_xy.y, og_settings.local_coord_precision)
        else:
            x_coord = round(mesonode.geometry.x, og_settings.lonlat_coord_precision)
            y_coord = round(mesonode.geometry.y, og_settings.lonlat_coord_precision)

        line = [mesonode.node_id, mesonode.zone_id, x_coord, y_coord, mesonode.macro_node_id, mesonode.macro_link_id,
                mesonode.activity_type, mesonode.is_boundary]
        writer.writerow(line)

    outfile.close()


def _outputMesoLinks(mesonet, mesolink_filepath, projection, encoding):
    outfile = getFileHandle(mesolink_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['link_id', 'from_node_id', 'to_node_id', 'dir_flag', 'length', 'lanes', 'capacity',
                     'free_speed', 'link_type_name', 'link_type', 'geometry', 'macro_link_id', 'macro_node_id',
                     'movement_id', 'mvmt_txt_id', 'ctrl_type', 'allowed_uses'])
    for _, mesolink in mesonet.link_dict.items():
        if projection:
            geometry_ = wkt.dumps(mesolink.geometry_xy, rounding_precision=og_settings.local_coord_precision)
        else:
            geometry_ = wkt.dumps(mesolink.geometry, rounding_precision=og_settings.lonlat_coord_precision)

        line = [mesolink.link_id, mesolink.from_node.node_id, mesolink.to_node.node_id, mesolink.dir_flag, mesolink.length,
                mesolink.lanes, mesolink.capacity, mesolink.free_speed, mesolink.link_type_name,mesolink.link_type,
                geometry_, mesolink.macro_link_id, mesolink.macro_node_id,  mesolink.movement_id, mesolink.mvmt_txt_id, mesolink.ctrl_type,
                ';'.join(mesolink.allowed_uses) if mesolink.allowed_uses else '']
        writer.writerow(line)

    outfile.close()


def _outputMicroNodes(micronet, micronode_filepath, projection, encoding):
    outfile = getFileHandle(micronode_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['node_id', 'zone_id', 'x_coord', 'y_coord', 'meso_link_id', 'lane_no', 'is_boundary'])
    for _, micronode in micronet.node_dict.items():
        if projection:
            x_coord = round(micronode.geometry_xy.x, og_settings.local_coord_precision)
            y_coord = round(micronode.geometry_xy.y, og_settings.local_coord_precision)
        else:
            x_coord = round(micronode.geometry.x, og_settings.lonlat_coord_precision)
            y_coord = round(micronode.geometry.y, og_settings.lonlat_coord_precision)

        line = [micronode.node_id, micronode.zone_id, x_coord, y_coord, micronode.mesolink.link_id, micronode.lane_no, micronode.is_boundary]
        writer.writerow(line)

    outfile.close()


def _outputMicroLinks(micronet, microlink_filepath, projection, encoding):
    outfile = getFileHandle(microlink_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['link_id', 'from_node_id', 'to_node_id', 'dir_flag', 'length', 'lanes', 'capacity',
                     'free_speed', 'link_type_name', 'link_type', 'geometry', 'macro_node_id', 'macro_link_id',
                     'meso_link_id', 'cell_type', 'additional_cost', 'lane_no', 'mvmt_txt_id', 'ctrl_type', 'allowed_uses'])
    for _, microlink in micronet.link_dict.items():
        if projection:
            geometry_ = wkt.dumps(microlink.geometry_xy, rounding_precision=og_settings.local_coord_precision)
        else:
            geometry_ = wkt.dumps(microlink.geometry, rounding_precision=og_settings.lonlat_coord_precision)

        line = [microlink.link_id, microlink.from_node.node_id, microlink.to_node.node_id, microlink.dir_flag,
                microlink.length, 1, microlink.mesolink.capacity, microlink.mesolink.free_speed,
                microlink.mesolink.link_type_name, microlink.mesolink.link_type, geometry_,
                microlink.mesolink.macro_node_id, microlink.mesolink.macro_link_id, microlink.mesolink.link_id,
                microlink.cell_type, 0, microlink.lane_no, microlink.mvmt_txt_id, microlink.ctrl_type,
                ';'.join(microlink.allowed_uses) if microlink.allowed_uses else '']
        writer.writerow(line)

    outfile.close()


def outputMesoNet(mesonet, output_folder, prefix, projection, encoding):
    mesonet_folder = os.path.join(output_folder, 'mesonet')
    if not os.path.isdir(mesonet_folder): os.mkdir(mesonet_folder)

    mesonode_filepath = os.path.join(mesonet_folder, f'{prefix}node.csv')
    _outputMesoNodes(mesonet, mesonode_filepath, projection, encoding)

    mesolink_filepath = os.path.join(mesonet_folder, f'{prefix}link.csv')
    _outputMesoLinks(mesonet, mesolink_filepath, projection, encoding)


def outputMicroNet(micronet, output_folder, prefix, projection, encoding):
    micronet_folder = os.path.join(output_folder, 'micronet')
    if not os.path.isdir(micronet_folder): os.mkdir(micronet_folder)

    micronode_filepath = os.path.join(micronet_folder, f'{prefix}node.csv')
    _outputMicroNodes(micronet, micronode_filepath, projection, encoding)

    microlink_filepath = os.path.join(micronet_folder, f'{prefix}link.csv')
    _outputMicroLinks(micronet, microlink_filepath, projection, encoding)



def _outputMacroNodes(network, macronode_filepath, projection, encoding):
    outfile = getFileHandle(macronode_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['name', 'node_id', 'osm_node_id', 'osm_highway', 'zone_id', 'ctrl_type', 'node_type', 'activity_type',
                     'is_boundary', 'x_coord', 'y_coord', 'intersection_id','poi_id','notes'])
    for node_id, node in network.node_dict.items():
        if projection:
            x_coord = round(node.geometry_xy.x, og_settings.local_coord_precision)
            y_coord = round(node.geometry_xy.y, og_settings.local_coord_precision)
        else:
            x_coord = round(node.geometry.x, og_settings.lonlat_coord_precision)
            y_coord = round(node.geometry.y, og_settings.lonlat_coord_precision)
        line = [node.name, node.node_id, node.osm_node_id, node.osm_highway, node.zone_id, node.ctrl_type, '', node.activity_type,
                node.is_boundary, x_coord, y_coord, node.intersection_id, node.poi_id, node.notes]
        writer.writerow(line)

    outfile.close()


def _outputMacroLinks(network, macrolink_filepath, projection, encoding):
    outfile = getFileHandle(macrolink_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['name', 'link_id', 'osm_way_id', 'from_node_id', 'to_node_id', 'dir_flag', 'length', 'lanes',
                     'free_speed', 'capacity', 'link_type_name', 'link_type', 'geometry','allowed_uses','from_biway',
                     'is_link', 'VDF_fftt1', 'VDF_cap1'])
    for link_id, link in network.link_dict.items():
        from_biway = 1 if link.from_bidirectional_way else 0
        is_link = 1 if link.is_link else 0
        if projection:
            geometry_ = wkt.dumps(link.geometry_xy, rounding_precision=og_settings.local_coord_precision)
        else:
            geometry_ = wkt.dumps(link.geometry, rounding_precision=og_settings.lonlat_coord_precision)
        line = [link.name, link.link_id, link.osm_way_id, link.from_node.node_id, link.to_node.node_id, link.dir_flag, link.length,
                link.lanes, link.free_speed, link.capacity, link.link_type_name, link.link_type, geometry_, ';'.join(link.allowed_uses) if link.allowed_uses else '',
                from_biway, is_link, link.VDF_fftt1, link.VDF_cap1]
        writer.writerow(line)

    outfile.close()


def outputMacroNet(network, output_folder, prefix, projection, encoding):
    macronet_folder = os.path.join(output_folder, 'macronet')
    if not os.path.isdir(macronet_folder): os.mkdir(macronet_folder)

    macronode_filepath = os.path.join(macronet_folder, f'{prefix}node.csv')
    _outputMacroNodes(network, macronode_filepath, projection, encoding)

    macrolink_filepath = os.path.join(macronet_folder, f'{prefix}link.csv')
    _outputMacroLinks(network, macrolink_filepath, projection, encoding)





from .multiresolution.build_mrnet import buildMultiResolutionNets
from .io.writefile import outputNetToCSV
from .osmnet.build_net import getNetFromFile
from .osmnet.complex_intersection import consolidateComplexIntersections
from .osmnet.enrich_net_info import generateNodeActivityInfo, generateLinkVDFInfo
from .osmnet.pois import connectPOIWithNet
from .io.load_from_csv import loadNetFromCSV

__all__ = [
    'buildMultiResolutionNets',
    'outputNetToCSV',
    'getNetFromFile',
    'consolidateComplexIntersections',
    'generateNodeActivityInfo',
    'generateLinkVDFInfo',
    'connectPOIWithNet',
    'loadNetFromCSV',
]





import copy

from ..networkclass.mesonet import MesoNode, MesoLink, MesoNetwork
from ..networkclass.micronet import MicroNode, MicroLink, MicroNetwork
from ..utils.util_geo import offsetLine
from .. import settings as og_settings
from shapely import geometry
import sys


class NetGenerator:
    def __init__(self, macronet, generate_micro_net, exclusive_bike_walk_lanes, length_of_cell, width_of_lane, num_nodes_for_ramp_alignment=8):
        self.macronet = macronet
        self.generate_micro_net = generate_micro_net
        self.exclusive_bike_walk_lanes = False
        self.length_of_cell = length_of_cell
        self.width_of_lane = width_of_lane
        self.num_nodes_for_ramp_alignment = num_nodes_for_ramp_alignment

        self.bike_lane_width = 0.5
        self.walk_lane_width = 0.5

        self.mesonet = MesoNetwork()
        self.micronet = MicroNetwork() if generate_micro_net else None

        self.number_of_expanded_mesonode = {}

    def getMultimoalUse(self, allowed_uses):
        if self.exclusive_bike_walk_lanes:
            if len(allowed_uses) <= 1:
                return {'mainlane_allowed_uses': allowed_uses, 'extra_bike': False, 'extra_walk': False}
            else:
                allowed_uses_set = set(allowed_uses).union({'auto','bike','walk'})
                if allowed_uses_set == {'auto','bike'}:
                    return {'mainlane_allowed_uses':['auto'], 'extra_bike':True, 'extra_walk':False}
                elif allowed_uses_set == {'auto','walk'}:
                    return {'mainlane_allowed_uses':['auto'], 'extra_bike':False, 'extra_walk':True}
                elif allowed_uses_set == {'bike','walk'}:
                    return {'mainlane_allowed_uses':['bike'], 'extra_bike':False, 'extra_walk':True}
                elif allowed_uses_set == {'auto','bike','walk'}:
                    return {'mainlane_allowed_uses':['auto'], 'extra_bike':True, 'extra_walk':True}
        else:
            return {'mainlane_allowed_uses':allowed_uses, 'extra_bike':False, 'extra_walk':False}

    def getLaneGeometry(self, original_geometry, lane_offset):
        if lane_offset < -1e-3 or lane_offset > 1e-3:

            lane_geometry_xy = original_geometry.offset_curve(distance=-1*lane_offset, join_style=2)
            if isinstance(lane_geometry_xy, geometry.MultiLineString):
                lane_geometry_xy = offsetLine(original_geometry, lane_offset)

            if lane_geometry_xy.is_empty:
                return self.getLaneGeometry(original_geometry, lane_offset*0.6)

            return lane_geometry_xy

        else:
            return copy.copy(original_geometry)

    # The remaining methods mirror the original implementation closely

    def createMicroNetForConnector(self, mesolink, ib_mesolink, ib_lane_index_start, ob_mesolink, ob_lane_index_start):
        max_micronode_id = self.micronet.max_node_id
        max_microlink_id = self.micronet.max_link_id

        for i in range(mesolink.lanes):
            start_micronode = ib_mesolink.micronode_list[ib_lane_index_start+i][-1]
            end_micronode = ob_mesolink.micronode_list[ob_lane_index_start+i][0]
            lane_geometry_xy = geometry.LineString([start_micronode.geometry_xy, end_micronode.geometry_xy])

            number_of_cells = max(1, round(lane_geometry_xy.length / self.length_of_cell))
            micronode_geometry_xy_list = [lane_geometry_xy.interpolate(i/number_of_cells, normalized=True) for i in range(1,number_of_cells)]

            mesolink.micronode_list.append([])
            mesolink.microlink_list.append([])
            last_micronode = start_micronode

            first_movement_cell = True

            for micronode_geometry_xy in micronode_geometry_xy_list:
                micronode = MicroNode(max_micronode_id)
                micronode.geometry_xy = micronode_geometry_xy
                micronode.geometry = self.macronet.GT.geo_to_latlon(micronode_geometry_xy)
                micronode.mesolink = mesolink
                micronode.lane_no = i + 1
                mesolink.micronode_list[-1].append(micronode)
                self.micronet.node_dict[micronode.node_id] = micronode
                max_micronode_id += 1

                microlink = MicroLink(max_microlink_id)
                microlink.from_node = last_micronode
                microlink.to_node = micronode
                microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                microlink.mesolink = mesolink
                microlink.cell_type = 1

                if first_movement_cell:
                    microlink.is_first_movement_cell = True
                    first_movement_cell = False

                mesolink.microlink_list[-1].append(microlink)
                self.micronet.link_dict[microlink.link_id] = microlink
                max_microlink_id += 1
                microlink.from_node.outgoing_link_list.append(microlink)
                microlink.to_node.incoming_link_list.append(microlink)

                last_micronode = micronode

            microlink = MicroLink(max_microlink_id)
            microlink.from_node = last_micronode
            microlink.to_node = end_micronode
            microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
            microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
            microlink.mesolink = mesolink
            microlink.cell_type = 1

            if first_movement_cell:
                microlink.is_first_movement_cell = True

            mesolink.microlink_list[-1].append(microlink)
            self.micronet.link_dict[microlink.link_id] = microlink
            max_microlink_id += 1
            microlink.from_node.outgoing_link_list.append(microlink)
            microlink.to_node.incoming_link_list.append(microlink)

        self.micronet.max_node_id = max_micronode_id
        self.micronet.max_link_id = max_microlink_id
    def createMicroNetForNormalLink(self, link):
        max_micronode_id = self.micronet.max_node_id
        max_microlink_id = self.micronet.max_link_id

        MultimoalUse = self.getMultimoalUse(link.allowed_uses)
        mainlane_allowed_uses, extra_bike, extra_walk = MultimoalUse['mainlane_allowed_uses'], MultimoalUse['extra_bike'], MultimoalUse['extra_walk']

        for mesolink in link.mesolink_list:
            original_number_of_lanes = mesolink.macrolink.lanes
            lane_changes_left = mesolink.lanes_change[0]
            num_of_lane_offset_between_left_most_and_central = -1 * (original_number_of_lanes / 2 - 0.5 + lane_changes_left)

            lane_geometry_xy_list = []
            extra_bike_geometry_xy, extra_walk_geometry_xy = None, None
            lane_offset = 0
            for i in range(mesolink.lanes):
                lane_offset = (num_of_lane_offset_between_left_most_and_central + i) * self.width_of_lane
                lane_geometry_xy_list.append(self.getLaneGeometry(mesolink.geometry_xy, lane_offset))
            if extra_bike and not extra_walk:
                bike_lane_offset = lane_offset + self.bike_lane_width
                extra_bike_geometry_xy = self.getLaneGeometry(mesolink.geometry_xy, bike_lane_offset)
            if not extra_bike and extra_walk:
                walk_lane_offset = lane_offset + self.walk_lane_width
                extra_walk_geometry_xy = self.getLaneGeometry(mesolink.geometry_xy, walk_lane_offset)
            if extra_bike and extra_walk:
                bike_lane_offset = lane_offset + self.bike_lane_width
                walk_lane_offset = bike_lane_offset + self.walk_lane_width
                extra_bike_geometry_xy = self.getLaneGeometry(mesolink.geometry_xy, bike_lane_offset)
                extra_walk_geometry_xy = self.getLaneGeometry(mesolink.geometry_xy, walk_lane_offset)

            # anchor: parallel-translate current mesolink lane centerlines to upstream mapped lanes (LEFT-aligned)
            try:
                meso_index = link.mesolink_list.index(mesolink)
                if meso_index > 0:
                    upstream_mesolink = link.mesolink_list[meso_index - 1]
                    if hasattr(upstream_mesolink, 'lane_geometry_xy_list') and upstream_mesolink.lane_geometry_xy_list:
                        up_left_index = upstream_mesolink.lanes_change[0]
                        down_left_index = mesolink.lanes_change[0]
                        min_left_index = min(up_left_index, down_left_index)
                        up_lane_index_start = up_left_index - min_left_index
                        down_lane_index_start = down_left_index - min_left_index

                        number_of_connecting_lanes = min(
                            upstream_mesolink.lanes - up_lane_index_start,
                            mesolink.lanes - down_lane_index_start
                        )

                        for j in range(number_of_connecting_lanes):
                            up_lane_index = up_lane_index_start + j
                            down_lane_index = down_lane_index_start + j
                            up_end_x, up_end_y = upstream_mesolink.lane_geometry_xy_list[up_lane_index].coords[-1]
                            down_start_x, down_start_y = lane_geometry_xy_list[down_lane_index].coords[0]
                            dx, dy = up_end_x - down_start_x, up_end_y - down_start_y
                            if abs(dx) > 1e-9 or abs(dy) > 1e-9:
                                shifted = [(x + dx, y + dy) for (x, y) in lane_geometry_xy_list[down_lane_index].coords]
                                lane_geometry_xy_list[down_lane_index] = geometry.LineString(shifted)
            except Exception:
                pass

            number_of_cells = max(1, round(mesolink.length / self.length_of_cell))
            micronode_geometry_xy_list = [[lane_geometry_xy.interpolate(i/number_of_cells, normalized=True) for i in range(number_of_cells+1)] for lane_geometry_xy in lane_geometry_xy_list]
            micronode_geometry_xy_bike = [extra_bike_geometry_xy.interpolate(i / number_of_cells, normalized=True) for i in range(number_of_cells + 1)] if extra_bike_geometry_xy is not None else None
            micronode_geometry_xy_walk = [extra_walk_geometry_xy.interpolate(i / number_of_cells, normalized=True) for i in range(number_of_cells + 1)] if extra_walk_geometry_xy is not None else None

            for i in range(mesolink.lanes):
                micronode_list_lane = []
                for micronode_geometry_xy in micronode_geometry_xy_list[i]:
                    micronode = MicroNode(max_micronode_id)
                    micronode.geometry_xy = micronode_geometry_xy
                    micronode.geometry = self.macronet.GT.geo_to_latlon(micronode_geometry_xy)
                    micronode.mesolink = mesolink
                    micronode.lane_no = i + 1
                    micronode_list_lane.append(micronode)
                    self.micronet.node_dict[micronode.node_id] = micronode
                    max_micronode_id += 1
                mesolink.micronode_list.append(micronode_list_lane)

            # persist current lane centerlines for downstream anchoring
            mesolink.lane_geometry_xy_list = lane_geometry_xy_list

            if extra_bike:
                for micronode_geometry_xy in micronode_geometry_xy_bike:
                    micronode = MicroNode(max_micronode_id)
                    micronode.geometry_xy = micronode_geometry_xy
                    micronode.geometry = self.macronet.GT.geo_to_latlon(micronode_geometry_xy)
                    micronode.mesolink = mesolink
                    mesolink.micronode_bike.append(micronode)
                    self.micronet.node_dict[micronode.node_id] = micronode
                    max_micronode_id += 1

            if extra_walk:
                for micronode_geometry_xy in micronode_geometry_xy_walk:
                    micronode = MicroNode(max_micronode_id)
                    micronode.geometry_xy = micronode_geometry_xy
                    micronode.geometry = self.macronet.GT.geo_to_latlon(micronode_geometry_xy)
                    micronode.mesolink = mesolink
                    mesolink.micronode_walk.append(micronode)
                    self.micronet.node_dict[micronode.node_id] = micronode
                    max_micronode_id += 1

        first_mesolink = link.mesolink_list[0]
        for micronode_list_lane in first_mesolink.micronode_list:
            micronode_list_lane[0].is_link_upstream_end_node = True
        if extra_bike: first_mesolink.micronode_bike[0].is_link_upstream_end_node = True
        if extra_walk: first_mesolink.micronode_walk[0].is_link_upstream_end_node = True
        last_mesolink = link.mesolink_list[-1]
        for micronode_list_lane in last_mesolink.micronode_list:
            micronode_list_lane[-1].is_link_downstream_end_node = True
        if extra_bike: last_mesolink.micronode_bike[-1].is_link_downstream_end_node = True
        if extra_walk: last_mesolink.micronode_walk[-1].is_link_downstream_end_node = True

        for i in range(len(link.mesolink_list)-1):
            upstream_mesolink = link.mesolink_list[i]
            downstream_mesolink = link.mesolink_list[i+1]

            up_index_of_left_most_lane_of_original_link = upstream_mesolink.lanes_change[0]
            down_index_of_left_most_lane_of_original_link = downstream_mesolink.lanes_change[0]
            min_left_most_lane_index = min(up_index_of_left_most_lane_of_original_link, down_index_of_left_most_lane_of_original_link)
            up_lane_index_start = up_index_of_left_most_lane_of_original_link - min_left_most_lane_index
            down_lane_index_start = down_index_of_left_most_lane_of_original_link - min_left_most_lane_index

            number_of_connecting_lanes = min(upstream_mesolink.lanes-up_lane_index_start,
                                             downstream_mesolink.lanes-down_lane_index_start)

            for j in range(number_of_connecting_lanes):
                up_lane_index = up_lane_index_start + j
                down_lane_index = down_lane_index_start + j
                up_micronode = upstream_mesolink.micronode_list[up_lane_index][-1]
                down_micronode = downstream_mesolink.micronode_list[down_lane_index][0]
                upstream_mesolink.micronode_list[up_lane_index][-1] = down_micronode
                del self.micronet.node_dict[up_micronode.node_id]

            if extra_bike:
                up_micronode = upstream_mesolink.micronode_bike[-1]
                down_micronode = downstream_mesolink.micronode_bike[0]
                upstream_mesolink.micronode_bike[-1] = down_micronode
                del self.micronet.node_dict[up_micronode.node_id]
            if extra_walk:
                up_micronode = upstream_mesolink.micronode_walk[-1]
                down_micronode = downstream_mesolink.micronode_walk[0]
                upstream_mesolink.micronode_walk = down_micronode
                del self.micronet.node_dict[up_micronode.node_id]

        for mesolink in link.mesolink_list:
            for i in range(mesolink.lanes):
                for j in range(len(mesolink.micronode_list[i])-1):
                    microlink = MicroLink(self.micronet.max_link_id)
                    self.micronet.max_link_id += 1
                    microlink.from_node = mesolink.micronode_list[i][j]
                    microlink.to_node = mesolink.micronode_list[i][j+1]
                    microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                    microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                    microlink.mesolink = mesolink
                    microlink.cell_type = 1
                    microlink.allowed_uses = mainlane_allowed_uses
                    self.micronet.link_dict[microlink.link_id] = microlink
                    microlink.from_node.outgoing_link_list.append(microlink)
                    microlink.to_node.incoming_link_list.append(microlink)

                if i <= mesolink.lanes - 2:
                    for j in range(len(mesolink.micronode_list[i])-1):
                        microlink = MicroLink(self.micronet.max_link_id)
                        self.micronet.max_link_id += 1
                        microlink.from_node = mesolink.micronode_list[i][j]
                        microlink.to_node = mesolink.micronode_list[i+1][j+1]
                        microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                        microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                        microlink.mesolink = mesolink
                        microlink.cell_type = 2
                        microlink.allowed_uses = mainlane_allowed_uses
                        self.micronet.link_dict[microlink.link_id] = microlink
                        microlink.from_node.outgoing_link_list.append(microlink)
                        microlink.to_node.incoming_link_list.append(microlink)

                if i >= 1:
                    for j in range(len(mesolink.micronode_list[i])-1):
                        microlink = MicroLink(self.micronet.max_link_id)
                        self.micronet.max_link_id += 1
                        microlink.from_node = mesolink.micronode_list[i][j]
                        microlink.to_node = mesolink.micronode_list[i-1][j+1]
                        microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                        microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                        microlink.mesolink = mesolink
                        microlink.cell_type = 2
                        microlink.allowed_uses = mainlane_allowed_uses
                        self.micronet.link_dict[microlink.link_id] = microlink
                        microlink.from_node.outgoing_link_list.append(microlink)
                        microlink.to_node.incoming_link_list.append(microlink)

            if extra_bike:
                for j in range(len(mesolink.micronode_bike) - 1):
                    microlink = MicroLink(self.micronet.max_link_id)
                    self.micronet.max_link_id += 1
                    microlink.from_node = mesolink.micronode_bike[j]
                    microlink.to_node = mesolink.micronode_bike[j + 1]
                    microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                    microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                    microlink.mesolink = mesolink
                    microlink.cell_type = 1
                    microlink.allowed_uses = ['bike']
                    self.micronet.link_dict[microlink.link_id] = microlink
                    microlink.from_node.outgoing_link_list.append(microlink)
                    microlink.to_node.incoming_link_list.append(microlink)

            if extra_walk:
                for j in range(len(mesolink.micronode_walk) - 1):
                    microlink = MicroLink(self.micronet.max_link_id)
                    self.micronet.max_link_id += 1
                    microlink.from_node = mesolink.micronode_walk[j]
                    microlink.to_node = mesolink.micronode_walk[j + 1]
                    microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                    microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                    microlink.mesolink = mesolink
                    microlink.cell_type = 1
                    microlink.allowed_uses = ['walk']
                    self.micronet.link_dict[microlink.link_id] = microlink
                    microlink.from_node.outgoing_link_list.append(microlink)
                    microlink.to_node.incoming_link_list.append(microlink)

        self.micronet.max_node_id = max(self.micronet.node_dict.keys()) + 1 if self.micronet.node_dict else 0
        self.micronet.max_link_id = max(self.micronet.link_dict.keys()) + 1 if self.micronet.link_dict else 0

    def createMesoNodeForCentriod(self):
        for node_id, node in self.macronet.node_dict.items():
            if node.is_centroid:
                if node not in self.number_of_expanded_mesonode.keys():
                    self.number_of_expanded_mesonode[node] = 0
                number_of_expanded_mesonode = self.number_of_expanded_mesonode[node]
                self.number_of_expanded_mesonode[node] += 1
                mesonode = MesoNode(node.node_id * 100 + number_of_expanded_mesonode)
                mesonode.geometry = node.geometry
                mesonode.geometry_xy = node.geometry_xy
                mesonode.macronode = node
                node.centroid_mesonode = mesonode
                self.mesonet.node_dict[mesonode.node_id] = mesonode

    def createNormalLinks(self):
        if og_settings.verbose:
            print('  generating normal meso links... (mrnet)')

        max_mesolink_id = self.mesonet.max_link_id

        for _, link in self.macronet.link_dict.items():
            macro_from_node = link.from_node
            if macro_from_node.is_centroid:
                upstream_node = macro_from_node.centroid_meso_node
            else:
                if macro_from_node not in self.number_of_expanded_mesonode.keys():
                    self.number_of_expanded_mesonode[macro_from_node] = 0
                number_of_expanded_mesonode = self.number_of_expanded_mesonode[macro_from_node]
                self.number_of_expanded_mesonode[macro_from_node] += 1

                upstream_node = MesoNode(macro_from_node.node_id * 100 + number_of_expanded_mesonode)
                upstream_node.geometry = geometry.Point(link.cutted_geometry_list[0].coords[0])
                upstream_node.geometry_xy = geometry.Point(link.cutted_geometry_xy_list[0].coords[0])
                upstream_node.macronode = macro_from_node

                self.mesonet.node_dict[upstream_node.node_id] = upstream_node

            cutted_number_of_segments = len(link.cutted_lanes_list)
            macro_to_node = link.to_node
            for section_no in range(cutted_number_of_segments):
                if macro_to_node.is_centroid and section_no == cutted_number_of_segments - 1:
                    downstream_node = macro_to_node.centroid_meso_node
                else:
                    if macro_to_node not in self.number_of_expanded_mesonode.keys():
                        self.number_of_expanded_mesonode[macro_to_node] = 0
                    number_of_expanded_mesonode = self.number_of_expanded_mesonode[macro_to_node]
                    self.number_of_expanded_mesonode[macro_to_node] += 1

                    downstream_node = MesoNode(macro_to_node.node_id * 100 + number_of_expanded_mesonode)
                    downstream_node.geometry = geometry.Point(link.cutted_geometry_list[section_no].coords[-1])
                    downstream_node.geometry_xy = geometry.Point(link.cutted_geometry_xy_list[section_no].coords[-1])
                    if section_no == cutted_number_of_segments - 1:
                        downstream_node.macronode = macro_to_node
                    else:
                        downstream_node.macrolink = link

                    self.mesonet.node_dict[downstream_node.node_id] = downstream_node

                mesolink = MesoLink(max_mesolink_id)
                mesolink.from_node = upstream_node
                mesolink.to_node = downstream_node
                mesolink.lanes = link.cutted_lanes_list[section_no]
                mesolink.lanes_change = link.cutted_lanes_change_list[section_no]
                mesolink.geometry = link.cutted_geometry_list[section_no]
                mesolink.geometry_xy = link.cutted_geometry_xy_list[section_no]
                mesolink.macrolink = link

                link.mesolink_list.append(mesolink)
                upstream_node.outgoing_link_list.append(mesolink)
                downstream_node.incoming_link_list.append(mesolink)

                self.mesonet.link_dict[mesolink.link_id] = mesolink
                max_mesolink_id += 1
                upstream_node = downstream_node

            if self.generate_micro_net:
                self.createMicroNetForNormalLink(link)

        self.mesonet.max_link_id = max_mesolink_id

    def connectMesoLinksMVMT(self):
        if og_settings.verbose:
            print('  generating movement meso links... (mrnet)')

        max_mesolink_id = self.mesonet.max_link_id

        for _, macronode in self.macronet.node_dict.items():
            for mvmt in macronode.movement_list:
                ib_link, ob_link = mvmt.ib_link, mvmt.ob_link
                ib_mesolink = ib_link.mesolink_list[-1]
                ob_mesolink = ob_link.mesolink_list[0]

                if macronode.movement_link_needed:
                    mesolink = MesoLink(max_mesolink_id)
                    mesolink.from_node = ib_mesolink.to_node
                    mesolink.to_node = ob_mesolink.from_node
                    mesolink.lanes = mvmt.lanes
                    mesolink.isconnector = True
                    mesolink.movement = mvmt
                    mesolink.macronode = macronode

                    mesolink.geometry = geometry.LineString([ib_mesolink.geometry.coords[-1], ob_mesolink.geometry.coords[0]])
                    mesolink.geometry_xy = geometry.LineString([ib_mesolink.geometry_xy.coords[-1], ob_mesolink.geometry_xy.coords[0]])

                    self.mesonet.link_dict[mesolink.link_id] = mesolink
                    max_mesolink_id += 1

                    mesolink.from_node.outgoing_link_list.append(mesolink)
                    mesolink.to_node.incoming_link_list.append(mesolink)
                    if self.generate_micro_net:
                        self.createMicroNetForConnector(mesolink, ib_mesolink, mvmt.start_ib_lane_seq_no, ob_mesolink, mvmt.start_ob_lane_seq_no)
                        # Group-wise alignment: off ramp vs on ramp with gradual transition
                        try:
                            affected = set()
                            # Off ramp (4->3+1): translate downstream first 6 nodes
                            if ib_mesolink.lanes >= ob_mesolink.lanes:
                                num_nodes_to_shift = min(self.num_nodes_for_ramp_alignment, len(ob_mesolink.micronode_list[0]) if ob_mesolink.micronode_list else 0)
                                for i_lane in range(mvmt.lanes):
                                    ib_idx = mvmt.start_ib_lane_seq_no + i_lane
                                    ob_idx = mvmt.start_ob_lane_seq_no + i_lane
                                    if ob_idx < 0 or ob_idx >= len(ob_mesolink.micronode_list):
                                        continue
                                    if ib_idx < 0 or ib_idx >= len(ib_mesolink.micronode_list):
                                        continue
                                    ib_end_node = ib_mesolink.micronode_list[ib_idx][-1]
                                    ob_lane_nodes = ob_mesolink.micronode_list[ob_idx]
                                    if not ob_lane_nodes:
                                        continue
                                    dx = ib_end_node.geometry_xy.x - ob_lane_nodes[0].geometry_xy.x
                                    dy = ib_end_node.geometry_xy.y - ob_lane_nodes[0].geometry_xy.y
                                    if abs(dx) < 1e-9 and abs(dy) < 1e-9:
                                        continue
                                    # Gradual transition: first 6 nodes with decreasing ratio
                                    for idx in range(num_nodes_to_shift):
                                        if idx < len(ob_lane_nodes):
                                            n = ob_lane_nodes[idx]
                                            ratio = 1.0 - (idx / num_nodes_to_shift)
                                            n.geometry_xy = geometry.Point((n.geometry_xy.x + dx * ratio, n.geometry_xy.y + dy * ratio))
                                            n.geometry = self.macronet.GT.geo_to_latlon(n.geometry_xy)
                                            for lk in n.outgoing_link_list:
                                                affected.add(lk.link_id)
                                            for lk in n.incoming_link_list:
                                                affected.add(lk.link_id)
                            # On ramp (3+1->4): translate upstream last 6 nodes per-lane
                            else:
                                num_nodes_to_shift = min(self.num_nodes_for_ramp_alignment, len(ib_mesolink.micronode_list[0]) if ib_mesolink.micronode_list else 0)
                                for i_lane in range(mvmt.lanes):
                                    ib_lane_idx = mvmt.start_ib_lane_seq_no + i_lane
                                    ob_lane_idx = mvmt.start_ob_lane_seq_no + i_lane
                                    if ib_lane_idx < 0 or ib_lane_idx >= len(ib_mesolink.micronode_list):
                                        continue
                                    if ob_lane_idx < 0 or ob_lane_idx >= len(ob_mesolink.micronode_list):
                                        continue
                                    ib_lane_end = ib_mesolink.micronode_list[ib_lane_idx][-1]
                                    ob_lane_start = ob_mesolink.micronode_list[ob_lane_idx][0]
                                    lane_dx = ib_lane_end.geometry_xy.x - ob_lane_start.geometry_xy.x
                                    lane_dy = ib_lane_end.geometry_xy.y - ob_lane_start.geometry_xy.y
                                    if abs(lane_dx) < 1e-9 and abs(lane_dy) < 1e-9:
                                        continue
                                    lane_nodes = ib_mesolink.micronode_list[ib_lane_idx]
                                    lane_len = len(lane_nodes)
                                    for idx in range(num_nodes_to_shift):
                                        node_idx = lane_len - num_nodes_to_shift + idx
                                        if 0 <= node_idx < lane_len:
                                            n = lane_nodes[node_idx]
                                            ratio = (idx + 1) / num_nodes_to_shift
                                            n.geometry_xy = geometry.Point((n.geometry_xy.x - lane_dx * ratio, n.geometry_xy.y - lane_dy * ratio))
                                            n.geometry = self.macronet.GT.geo_to_latlon(n.geometry_xy)
                                            for lk in n.outgoing_link_list:
                                                affected.add(lk.link_id)
                                            for lk in n.incoming_link_list:
                                                affected.add(lk.link_id)
                            # Rebuild connector's intermediate nodes after alignment
                            # Both off ramp and on ramp need this, as endpoints have moved
                            if mesolink and mesolink.isconnector and hasattr(mesolink, 'micronode_list'):
                                for connector_lane_idx in range(len(mesolink.micronode_list)):
                                    if connector_lane_idx < mvmt.lanes:
                                        ob_lane_idx = mvmt.start_ob_lane_seq_no + connector_lane_idx
                                        ib_lane_idx = mvmt.start_ib_lane_seq_no + connector_lane_idx
                                        if (0 <= ob_lane_idx < len(ob_mesolink.micronode_list) and 
                                            0 <= ib_lane_idx < len(ib_mesolink.micronode_list)):
                                            connector_nodes = mesolink.micronode_list[connector_lane_idx]
                                            if len(connector_nodes) > 0:
                                                # Get new endpoint positions after alignment
                                                ib_end = ib_mesolink.micronode_list[ib_lane_idx][-1]
                                                ob_start = ob_mesolink.micronode_list[ob_lane_idx][0]
                                                # Rebuild connector geometry with new endpoints
                                                total_nodes = len(connector_nodes) + 2  # including start and end
                                                for idx, n in enumerate(connector_nodes):
                                                    # Interpolate between new start and end positions
                                                    t = (idx + 1) / total_nodes
                                                    new_x = ib_end.geometry_xy.x * (1 - t) + ob_start.geometry_xy.x * t
                                                    new_y = ib_end.geometry_xy.y * (1 - t) + ob_start.geometry_xy.y * t
                                                    n.geometry_xy = geometry.Point((new_x, new_y))
                                                    n.geometry = self.macronet.GT.geo_to_latlon(n.geometry_xy)
                                                    for lk in n.outgoing_link_list:
                                                        affected.add(lk.link_id)
                                                    for lk in n.incoming_link_list:
                                                        affected.add(lk.link_id)
                            for lid in affected:
                                ml = self.micronet.link_dict.get(lid)
                                if ml is not None:
                                    ml.geometry_xy = geometry.LineString([ml.from_node.geometry_xy, ml.to_node.geometry_xy])
                                    ml.geometry = geometry.LineString([ml.from_node.geometry, ml.to_node.geometry])
                        except Exception:
                            pass
                else:
                    if ib_link.downstream_is_target and not ob_link.upstream_is_target:
                        ib_mesolink_to_node = ib_mesolink.to_node
                        ob_mesolink_from_node = ob_mesolink.from_node
                        ob_mesolink.from_node = ib_mesolink_to_node
                        ob_mesolink.geometry = geometry.LineString([ib_mesolink.geometry.coords[-1]] + ob_mesolink.geometry.coords[1:])
                        ob_mesolink.geometry_xy = geometry.LineString([ib_mesolink.geometry_xy.coords[-1]] + ob_mesolink.geometry_xy.coords[1:])
                        del self.mesonet.node_dict[ob_mesolink_from_node.node_id]
                        if self.generate_micro_net:
                            for i in range(mvmt.lanes):
                                ib_lane_index = mvmt.start_ib_lane_seq_no + i
                                ob_lane_index = mvmt.start_ob_lane_seq_no + i
                                ib_mesolink_outgoing_micro_node = ib_mesolink.micronode_list[ib_lane_index][-1]
                                ob_mesolink_incoming_micro_node = ob_mesolink.micronode_list[ob_lane_index][0]
                                for microlink in ob_mesolink_incoming_micro_node.outgoing_link_list:
                                    microlink.from_node = ib_mesolink_outgoing_micro_node
                                del self.micronet.node_dict[ob_mesolink_incoming_micro_node.node_id]
                            # Group-wise alignment: off ramp vs on ramp
                            try:
                                affected = set()
                                if ib_mesolink.lanes >= ob_mesolink.lanes:
                                    num_nodes_to_shift = min(self.num_nodes_for_ramp_alignment, len(ob_mesolink.micronode_list[0]) if ob_mesolink.micronode_list else 0)
                                    for i_lane in range(mvmt.lanes):
                                        ib_idx = mvmt.start_ib_lane_seq_no + i_lane
                                        ob_idx = mvmt.start_ob_lane_seq_no + i_lane
                                        if ob_idx < 0 or ob_idx >= len(ob_mesolink.micronode_list):
                                            continue
                                        if ib_idx < 0 or ib_idx >= len(ib_mesolink.micronode_list):
                                            continue
                                        ib_end_node = ib_mesolink.micronode_list[ib_idx][-1]
                                        ob_lane_nodes = ob_mesolink.micronode_list[ob_idx]
                                        if not ob_lane_nodes:
                                            continue
                                        dx = ib_end_node.geometry_xy.x - ob_lane_nodes[0].geometry_xy.x
                                        dy = ib_end_node.geometry_xy.y - ob_lane_nodes[0].geometry_xy.y
                                        if abs(dx) < 1e-9 and abs(dy) < 1e-9:
                                            continue
                                        for idx in range(num_nodes_to_shift):
                                            if idx < len(ob_lane_nodes):
                                                n = ob_lane_nodes[idx]
                                                ratio = 1.0 - (idx / num_nodes_to_shift)
                                                n.geometry_xy = geometry.Point((n.geometry_xy.x + dx * ratio, n.geometry_xy.y + dy * ratio))
                                                n.geometry = self.macronet.GT.geo_to_latlon(n.geometry_xy)
                                                for lk in n.outgoing_link_list:
                                                    affected.add(lk.link_id)
                                                for lk in n.incoming_link_list:
                                                    affected.add(lk.link_id)
                                else:
                                    num_nodes_to_shift = min(self.num_nodes_for_ramp_alignment, len(ib_mesolink.micronode_list[0]) if ib_mesolink.micronode_list else 0)
                                    for i_lane in range(mvmt.lanes):
                                        ib_lane_idx = mvmt.start_ib_lane_seq_no + i_lane
                                        ob_lane_idx = mvmt.start_ob_lane_seq_no + i_lane
                                        if ib_lane_idx < 0 or ib_lane_idx >= len(ib_mesolink.micronode_list):
                                            continue
                                        if ob_lane_idx < 0 or ob_lane_idx >= len(ob_mesolink.micronode_list):
                                            continue
                                        ib_lane_end = ib_mesolink.micronode_list[ib_lane_idx][-1]
                                        ob_lane_start = ob_mesolink.micronode_list[ob_lane_idx][0]
                                        lane_dx = ib_lane_end.geometry_xy.x - ob_lane_start.geometry_xy.x
                                        lane_dy = ib_lane_end.geometry_xy.y - ob_lane_start.geometry_xy.y
                                        if abs(lane_dx) < 1e-9 and abs(lane_dy) < 1e-9:
                                            continue
                                        lane_nodes = ib_mesolink.micronode_list[ib_lane_idx]
                                        lane_len = len(lane_nodes)
                                        for idx in range(num_nodes_to_shift):
                                            node_idx = lane_len - num_nodes_to_shift + idx
                                            if 0 <= node_idx < lane_len:
                                                n = lane_nodes[node_idx]
                                                ratio = (idx + 1) / num_nodes_to_shift
                                                n.geometry_xy = geometry.Point((n.geometry_xy.x - lane_dx * ratio, n.geometry_xy.y - lane_dy * ratio))
                                                n.geometry = self.macronet.GT.geo_to_latlon(n.geometry_xy)
                                                for lk in n.outgoing_link_list:
                                                    affected.add(lk.link_id)
                                                for lk in n.incoming_link_list:
                                                    affected.add(lk.link_id)
                                for lid in affected:
                                    ml = self.micronet.link_dict.get(lid)
                                    if ml is not None:
                                        ml.geometry_xy = geometry.LineString([ml.from_node.geometry_xy, ml.to_node.geometry_xy])
                                        ml.geometry = geometry.LineString([ml.from_node.geometry, ml.to_node.geometry])
                            except Exception:
                                pass
                    elif not ib_link.downstream_is_target and ob_link.upstream_is_target:
                        ib_mesolink_to_node = ib_mesolink.to_node
                        ob_mesolink_from_node = ob_mesolink.from_node
                        ib_mesolink.to_node = ob_mesolink_from_node
                        ib_mesolink.geometry = geometry.LineString(ib_mesolink.geometry.coords[:-1] + [ob_mesolink.geometry.coords[0]])
                        ib_mesolink.geometry_xy = geometry.LineString(ib_mesolink.geometry_xy.coords[:-1] + [ob_mesolink.geometry_xy.coords[0]])
                        del self.mesonet.node_dict[ib_mesolink_to_node.node_id]
                        if self.generate_micro_net:
                            for i in range(mvmt.lanes):
                                ib_lane_index = mvmt.start_ib_lane_seq_no + i
                                ob_lane_index = mvmt.start_ob_lane_seq_no + i
                                ib_mesolink_outgoing_micro_node = ib_mesolink.micronode_list[ib_lane_index][-1]
                                ob_mesolink_incoming_micro_node = ob_mesolink.micronode_list[ob_lane_index][0]
                                for microlink in ib_mesolink_outgoing_micro_node.incoming_link_list:
                                    microlink.to_node = ob_mesolink_incoming_micro_node
                                del self.micronet.node_dict[ib_mesolink_outgoing_micro_node.node_id]
                            # Group-wise alignment: off ramp vs on ramp
                            try:
                                affected = set()
                                if ib_mesolink.lanes >= ob_mesolink.lanes:
                                    num_nodes_to_shift = min(self.num_nodes_for_ramp_alignment, len(ob_mesolink.micronode_list[0]) if ob_mesolink.micronode_list else 0)
                                    for i_lane in range(mvmt.lanes):
                                        ib_idx = mvmt.start_ib_lane_seq_no + i_lane
                                        ob_idx = mvmt.start_ob_lane_seq_no + i_lane
                                        if ob_idx < 0 or ob_idx >= len(ob_mesolink.micronode_list):
                                            continue
                                        if ib_idx < 0 or ib_idx >= len(ib_mesolink.micronode_list):
                                            continue
                                        ib_end_node = ib_mesolink.micronode_list[ib_idx][-1]
                                        ob_lane_nodes = ob_mesolink.micronode_list[ob_idx]
                                        if not ob_lane_nodes:
                                            continue
                                        dx = ib_end_node.geometry_xy.x - ob_lane_nodes[0].geometry_xy.x
                                        dy = ib_end_node.geometry_xy.y - ob_lane_nodes[0].geometry_xy.y
                                        if abs(dx) < 1e-9 and abs(dy) < 1e-9:
                                            continue
                                        for idx in range(num_nodes_to_shift):
                                            if idx < len(ob_lane_nodes):
                                                n = ob_lane_nodes[idx]
                                                ratio = 1.0 - (idx / num_nodes_to_shift)
                                                n.geometry_xy = geometry.Point((n.geometry_xy.x + dx * ratio, n.geometry_xy.y + dy * ratio))
                                                n.geometry = self.macronet.GT.geo_to_latlon(n.geometry_xy)
                                                for lk in n.outgoing_link_list:
                                                    affected.add(lk.link_id)
                                                for lk in n.incoming_link_list:
                                                    affected.add(lk.link_id)
                                else:
                                    num_nodes_to_shift = min(self.num_nodes_for_ramp_alignment, len(ib_mesolink.micronode_list[0]) if ib_mesolink.micronode_list else 0)
                                    for i_lane in range(mvmt.lanes):
                                        ib_lane_idx = mvmt.start_ib_lane_seq_no + i_lane
                                        ob_lane_idx = mvmt.start_ob_lane_seq_no + i_lane
                                        if ib_lane_idx < 0 or ib_lane_idx >= len(ib_mesolink.micronode_list):
                                            continue
                                        if ob_lane_idx < 0 or ob_lane_idx >= len(ob_mesolink.micronode_list):
                                            continue
                                        ib_lane_end = ib_mesolink.micronode_list[ib_lane_idx][-1]
                                        ob_lane_start = ob_mesolink.micronode_list[ob_lane_idx][0]
                                        lane_dx = ib_lane_end.geometry_xy.x - ob_lane_start.geometry_xy.x
                                        lane_dy = ib_lane_end.geometry_xy.y - ob_lane_start.geometry_xy.y
                                        if abs(lane_dx) < 1e-9 and abs(lane_dy) < 1e-9:
                                            continue
                                        lane_nodes = ib_mesolink.micronode_list[ib_lane_idx]
                                        lane_len = len(lane_nodes)
                                        for idx in range(num_nodes_to_shift):
                                            node_idx = lane_len - num_nodes_to_shift + idx
                                            if 0 <= node_idx < lane_len:
                                                n = lane_nodes[node_idx]
                                                ratio = (idx + 1) / num_nodes_to_shift
                                                n.geometry_xy = geometry.Point((n.geometry_xy.x - lane_dx * ratio, n.geometry_xy.y - lane_dy * ratio))
                                                n.geometry = self.macronet.GT.geo_to_latlon(n.geometry_xy)
                                                for lk in n.outgoing_link_list:
                                                    affected.add(lk.link_id)
                                                for lk in n.incoming_link_list:
                                                    affected.add(lk.link_id)
                                for lid in affected:
                                    ml = self.micronet.link_dict.get(lid)
                                    if ml is not None:
                                        ml.geometry_xy = geometry.LineString([ml.from_node.geometry_xy, ml.to_node.geometry_xy])
                                        ml.geometry = geometry.LineString([ml.from_node.geometry, ml.to_node.geometry])
                            except Exception:
                                pass
                    else:
                        sys.exit('Target link defintion error')

        self.mesonet.max_link_id = max_mesolink_id

    def generateNet(self):
        self.createMesoNodeForCentriod()
        self.createNormalLinks()
        self.connectMesoLinksMVMT()









import mrnet4gmns as mr

net = mr.loadNetFromCSV(node_file='net_full/node.csv', link_file='net_full/link.csv')
mr.buildMultiResolutionNets(net, width_of_lane=3.50, length_of_cell=7.0, num_nodes_for_ramp_alignment=10)
mr.outputNetToCSV(net, output_folder='mrn_full')



import math
from shapely import geometry

_indent = 8.0


def getMovementDescription(ib_link, ob_link):
    ib_coords = list(ib_link.geometry_xy.coords)
    ob_coords = list(ob_link.geometry_xy.coords)
    
    # Direction should be based on link END (approaching junction), not start
    # Use local direction at link end (last few points) to handle curved links
    ib_end = ib_coords[-1]
    if len(ib_coords) >= 3:
        # Use direction from third-to-last point to last point
        ib_direction_start = ib_coords[-3]
    else:
        ib_direction_start = ib_coords[0]

    angle_ib = math.atan2(ib_end[1] - ib_direction_start[1], ib_end[0] - ib_direction_start[0])
    if -0.75 * math.pi <= angle_ib < -0.25 * math.pi:
        direction = 'SB'
    elif -0.25 * math.pi <= angle_ib < 0.25 * math.pi:
        direction = 'EB'
    elif 0.25 * math.pi <= angle_ib < 0.75 * math.pi:
        direction = 'NB'
    else:
        direction = 'WB'

    # For turn angle, use local directions at the junction
    # ib_link: direction at end (last few points)
    ib_end = ib_coords[-1]
    if len(ib_coords) >= 3:
        ib_end_direction_point = ib_coords[-3]
    else:
        ib_end_direction_point = ib_coords[0]
    
    # ob_link: direction at start (first few points)
    ob_start = ob_coords[0]
    if len(ob_coords) >= 3:
        ob_start_direction_point = ob_coords[2]
    else:
        ob_start_direction_point = ob_coords[-1]
    
    angle_ib_end = math.atan2(ib_end[1] - ib_end_direction_point[1], ib_end[0] - ib_end_direction_point[0])
    angle_ob_start = math.atan2(ob_start_direction_point[1] - ob_start[1], ob_start_direction_point[0] - ob_start[0])
    angle = angle_ob_start - angle_ib_end
    if angle < -1 * math.pi:
        angle += 2 * math.pi
    if angle > math.pi:
        angle -= 2 * math.pi

    if -0.25 * math.pi <= angle <= 0.25 * math.pi:
        mvmt = 'T'
        mvmt_type = 'thru'
    elif angle < -0.25 * math.pi:
        mvmt = 'R'
        mvmt_type = 'right'
    elif angle <= 0.75 * math.pi:
        mvmt = 'L'
        mvmt_type = 'left'
    else:
        mvmt = 'U'
        mvmt_type = 'uturn'

    mvmt_txt_id = direction + mvmt
    return mvmt_txt_id, mvmt_type


def getMovementGeometry(ib_link, ob_link):
    ib_geometry_xy = ib_link.geometry_xy
    ib_indent = _indent if ib_geometry_xy.length > _indent else ib_geometry_xy.length / 2
    ib_point = ib_geometry_xy.interpolate(-1 * ib_indent)

    ob_geometry_xy = ob_link.geometry_xy
    ob_indent = _indent if ob_geometry_xy.length > _indent else ob_geometry_xy.length / 2
    ob_point = ob_geometry_xy.interpolate(ob_indent)

    geometry_xy = geometry.LineString([ib_point, ob_point])
    return geometry_xy






from .autoconintd import CAutoConnectorIntD
from .autoconm import CAutoConnectorM
from .util_mvmt import getMovementDescription, getMovementGeometry
from ..networkclass.macronet import Movement
from .. import settings as og_settings


def _generateMovementsForOneNode(node, max_movement_id, GT):
    number_of_incoming_links = len(node.incoming_link_list)
    number_of_outgoing_links = len(node.outgoing_link_list)

    if number_of_incoming_links == 0 or number_of_outgoing_links == 0: return

    movement_id = max_movement_id

    if number_of_outgoing_links == 1:
        ob_link = node.outgoing_link_list[0]
        ib_link_list = []
        for ib_link in node.incoming_link_list:
            if ib_link.from_node.node_id != ob_link.to_node.node_id:
                ib_link_list.append(ib_link)

        if len(ib_link_list) == 0: return

        CAutoConnectorM.ob_link, CAutoConnectorM.ib_link_list = ob_link, ib_link_list
        connection_list = CAutoConnectorM.buildConnector()

        for ib_link_no, ib_link in enumerate(ib_link_list):
            ib_lane_index_start = connection_list[ib_link_no][0][0]
            ib_lane_index_end = connection_list[ib_link_no][0][1]
            ob_lane_index_start = connection_list[ib_link_no][1][0]
            ob_lane_index_end = connection_list[ib_link_no][1][1]
            number_of_lanes = ib_lane_index_end - ib_lane_index_start + 1

            mvmt = Movement(movement_id)
            mvmt.node = node
            mvmt.ib_link = ib_link
            mvmt.ob_link = ob_link
            mvmt.start_ib_lane_seq_no, mvmt.end_ib_lane_seq_no = ib_lane_index_start, ib_lane_index_end
            mvmt.start_ob_lane_seq_no, mvmt.end_ob_lane_seq_no = ob_lane_index_start, ob_lane_index_end
            mvmt.start_ib_lane, mvmt.end_ib_lane = ib_link.outgoing_lane_indices[ib_lane_index_start], ib_link.outgoing_lane_indices[ib_lane_index_end]
            mvmt.start_ob_lane, mvmt.end_ob_lane = ob_link.incoming_lane_indices[ob_lane_index_start], ob_link.incoming_lane_indices[ob_lane_index_end]
            mvmt.lanes = number_of_lanes
            mvmt_txt_id, mvmt_type = getMovementDescription(ib_link, ob_link)
            mvmt.mvmt_txt_id = mvmt_txt_id
            mvmt.type = mvmt_type
            mvmt.geometry_xy = getMovementGeometry(ib_link, ob_link)
            mvmt.geometry = GT.geo_to_latlon(mvmt.geometry_xy)
            mvmt.generated_by_mrnet = True

            if node.ctrl_type == 1: mvmt.ctrl_type = 'signal'
            mvmt.allowed_uses = ib_link.allowed_uses
            node.movement_list.append(mvmt)
            movement_id += 1
    else:
        for ib_link in node.incoming_link_list:
            ob_link_list = []
            for ob_link in node.outgoing_link_list:
                if ib_link.from_node.node_id != ob_link.to_node.node_id:
                    ob_link_list.append(ob_link)

            if len(ob_link_list) == 0: continue

            CAutoConnectorIntD.ib_link, CAutoConnectorIntD.ob_link_list = ib_link, ob_link_list
            connection_list = CAutoConnectorIntD.buildConnector()

            for ob_link_no, ob_link in enumerate(ob_link_list):
                ib_lane_index_start = connection_list[ob_link_no][0][0]
                ib_lane_index_end = connection_list[ob_link_no][0][1]
                ob_lane_index_start = connection_list[ob_link_no][1][0]
                ob_lane_index_end = connection_list[ob_link_no][1][1]
                number_of_lanes = ib_lane_index_end - ib_lane_index_start + 1

                mvmt = Movement(movement_id)
                mvmt.node = node
                mvmt.ib_link = ib_link
                mvmt.ob_link = ob_link
                mvmt.start_ib_lane_seq_no, mvmt.end_ib_lane_seq_no = ib_lane_index_start, ib_lane_index_end
                mvmt.start_ob_lane_seq_no, mvmt.end_ob_lane_seq_no = ob_lane_index_start, ob_lane_index_end
                mvmt.start_ib_lane, mvmt.end_ib_lane = ib_link.outgoing_lane_indices[ib_lane_index_start], ib_link.outgoing_lane_indices[ib_lane_index_end]
                mvmt.start_ob_lane, mvmt.end_ob_lane = ob_link.incoming_lane_indices[ob_lane_index_start], ob_link.incoming_lane_indices[ob_lane_index_end]
                mvmt.lanes = number_of_lanes
                mvmt_txt_id, mvmt_type = getMovementDescription(ib_link, ob_link)
                mvmt.mvmt_txt_id = mvmt_txt_id
                mvmt.type = mvmt_type
                mvmt.geometry_xy = getMovementGeometry(ib_link, ob_link)
                mvmt.geometry = GT.geo_to_latlon(mvmt.geometry_xy)
                mvmt.generated_by_mrnet = True

                if node.ctrl_type == 1: mvmt.ctrl_type = 'signal'
                mvmt.allowed_uses = ib_link.allowed_uses
                node.movement_list.append(mvmt)
                movement_id += 1


def validateUserInputMovements(network):
    for movement in network.user_input_movement_list:
        ib_link, ob_link = movement.ib_link, movement.ob_link

        if movement.start_ib_lane not in ib_link.outgoing_lane_indices:
            print(f'WARNING: start_ib_lane of movement {movement.movement_id} does not belong to ib_link {ib_link.link_id} outgoing lane indices {ib_link.outgoing_lane_indices}')
            continue
        start_ib_lane_seq_no = ib_link.outgoing_lane_indices.index(movement.start_ib_lane)

        if movement.end_ib_lane not in ib_link.outgoing_lane_indices:
            print(f'WARNING: end_ib_lane of movement {movement.movement_id} does not belong to ib_link {ib_link.link_id} outgoing lane indices {ib_link.outgoing_lane_indices}')
            continue
        end_ib_lane_seq_no = ib_link.outgoing_lane_indices.index(movement.end_ib_lane)

        if movement.start_ob_lane not in ob_link.incoming_lane_indices:
            print(f'WARNING: start_ob_lane of movement {movement.movement_id} does not belong to ob_link {ob_link.link_id} incoming lane indices {ob_link.incoming_lane_indices}')
            continue
        start_ob_lane_seq_no = ob_link.incoming_lane_indices.index(movement.start_ob_lane)

        if movement.end_ob_lane not in ob_link.incoming_lane_indices:
            print(f'WARNING: end_ob_lane of movement {movement.movement_id} does not belong to ob_link {ob_link.link_id} incoming lane indices {ob_link.incoming_lane_indices}')
            continue
        end_ob_lane_seq_no = ob_link.incoming_lane_indices.index(movement.end_ob_lane)

        ib_lanes, ob_lanes = end_ib_lane_seq_no - start_ib_lane_seq_no + 1, end_ob_lane_seq_no - start_ob_lane_seq_no + 1

        if ib_lanes < ob_lanes:
            end_ob_lane_seq_no -= (ob_lanes - ib_lanes)
            end_ob_lane = ob_link.incoming_lane_indices[end_ob_lane_seq_no]
            print(f'WARNING: movement {movement.movement_id} - number of ib lanes ({ib_lanes}) is less than that of ob lanes ({ob_lanes}), end_ob_lane is changed to {end_ob_lane}')
            movement.lanes = ib_lanes
        elif ib_lanes > ob_lanes:
            end_ib_lane_seq_no -= (ib_lanes - ob_lanes)
            end_ib_lane = ib_link.outgoing_lane_indices[end_ib_lane_seq_no]
            print(f'WARNING: movement {movement.movement_id} - number of ib lanes ({ib_lanes}) is more than that of ob lanes ({ob_lanes}), end_ib_lane is changed to {end_ib_lane}')
            movement.lanes = ob_lanes
        else:
            movement.lanes = ib_lanes

        movement.end_ib_lane, movement.start_ib_lane_seq_no, movement.end_ib_lane_seq_no = movement.end_ib_lane, start_ib_lane_seq_no, end_ib_lane_seq_no
        movement.end_ob_lane, movement.start_ob_lane_seq_no, movement.end_ob_lane_seq_no = movement.end_ob_lane, start_ob_lane_seq_no, end_ob_lane_seq_no
        movement.node.movement_list.append(movement)


def generateMovements(network):
    if og_settings.verbose:
        print('Generating Movements (mrnet)')

    if not network.complete_link_lane_info:
        print('WARNING: Movement generation skipped: some links are missing lanes info.')
        return

    for _, link in network.link_dict.items():
        link.linkLaneListFromSegment()

    validateUserInputMovements(network)

    max_movement_id = network.max_movement_id
    for node_id, node in network.node_dict.items():
        if node.movement_list:
            continue
        _generateMovementsForOneNode(node, max_movement_id, network.GT)
        max_movement_id += len(node.movement_list)

    network.max_movement_id = max_movement_id






from ..networkclass.macronet import Node, Link, Network
from .read_from_osm import readOSMFile
from .osmclasses import OSMNode
from .wayfilters import *
from .check_args import checkArgs_getNetFromFile
from .combine_links import combineShortLinks
from ..utils.util import getLogger
from ..utils.util_geo import offsetLine
from .. import settings as og_settings
from shapely import geometry


def _createNode(network, osmnode):
    node = Node(network.max_node_id)
    node.name = None
    node.osm_node_id = osmnode['id']
    node.geometry = osmnode['geom']
    node.geometry_xy = osmnode['geom_xy']
    network.node_dict[node.node_id] = node
    network.max_node_id += 1
    return node


def _getNode(network, osmnode_dict, osm_id_to_node):
    if osmnode_dict['id'] in osm_id_to_node:
        return osm_id_to_node[osmnode_dict['id']]
    n = _createNode(network, osmnode_dict)
    osm_id_to_node[osmnode_dict['id']] = n
    return n


def _default_lanes(highway, default_lanes):
    if default_lanes is False or default_lanes is None:
        return None
    table = og_settings.default_lanes_dict
    key = highway if highway in table else 'primary'
    return table[key]


def _default_speed(highway, default_speed):
    if default_speed is False or default_speed is None:
        return None
    table = og_settings.default_speed_dict
    key = highway if highway in table else 'primary'
    return table[key]


def _default_capacity(highway, default_capacity):
    if default_capacity is False or default_capacity is None:
        return None
    table = og_settings.default_capacity_dict
    key = highway if highway in table else 'primary'
    return table[key]


def _createNodeOnBoundary(node_in, node_outside, network):
    line = network.bounds.intersection(geometry.LineString([node_in.geometry,node_outside.geometry]))
    lon, lat = line.coords[1]
    geometry_lonlat = geometry.Point((round(lon,og_settings.lonlat_coord_precision),round(lat,og_settings.lonlat_coord_precision)))
    boundary_osm_node = OSMNode('', '', geometry_lonlat, True, '', '')
    boundary_osm_node.geometry_xy = network.GT.geo_from_latlon(geometry_lonlat)
    boundary_osm_node.is_crossing = True
    boundary_osm_node.notes = 'boundary node created by mrnet'
    return boundary_osm_node


def _getSegmentNodeList(way, segment_no, network):
    m_segment_node_list = way.segment_node_list[segment_no]
    if way.is_reversed: m_segment_node_list = list(reversed(m_segment_node_list))
    number_of_nodes = len(m_segment_node_list)

    m_segment_node_list_group = []

    if m_segment_node_list[0].in_region:
        idx_first_outside = -1
        for idx, node in enumerate(m_segment_node_list):
            if not node.in_region:
                idx_first_outside = idx
                break

        if idx_first_outside == -1:
            m_segment_node_list_group.append(m_segment_node_list)
            return m_segment_node_list_group
        else:
            new_node = _createNodeOnBoundary(m_segment_node_list[idx_first_outside-1],m_segment_node_list[idx_first_outside], network)
            m_segment_node_list_group.append(m_segment_node_list[:idx_first_outside] + [new_node])

    if m_segment_node_list[-1].in_region:
        idx_last_outside = -1
        for idx in range(number_of_nodes-2,-1,-1):
            if not m_segment_node_list[idx].in_region:
                idx_last_outside = idx
                break
        new_node = _createNodeOnBoundary(m_segment_node_list[idx_last_outside+1],m_segment_node_list[idx_last_outside], network)
        m_segment_node_list_group.append([new_node] + m_segment_node_list[idx_last_outside+1:])

    return m_segment_node_list_group


def _createNodeFromOSMNode(network, osmnode):
    if osmnode.node is None:
        node = Node(network.max_node_id)
        node.name = osmnode.name
        node.osm_node_id = osmnode.osm_node_id
        node.osm_highway = osmnode.osm_highway
        node.ctrl_type = osmnode.ctrl_type
        node.geometry = osmnode.geometry
        node.geometry_xy = osmnode.geometry_xy
        node.notes = osmnode.notes
        osmnode.node = node
        network.node_dict[node.node_id] = node
        network.max_node_id += 1


def _createNodesAndLinks(network, link_way_list):
    if og_settings.verbose:
        print('    generating nodes and links (mrnet)')

    link_dict = {}
    max_link_id = network.max_link_id
    for way in link_way_list:
        if way.is_pure_cycle: continue
        way.getNodeListForSegments()
        for segment_no in range(way.number_of_segments):
            m_segment_node_list_group = _getSegmentNodeList(way, segment_no, network) if network.bounds is not None else [way.segment_node_list[segment_no]]
            for m_segment_node_list in m_segment_node_list_group:
                if len(m_segment_node_list) < 2: continue
                _createNodeFromOSMNode(network, m_segment_node_list[0])
                _createNodeFromOSMNode(network, m_segment_node_list[-1])

                link = Link(max_link_id)
                link.buildFromOSMWay(way, 1, m_segment_node_list, network.default_lanes, network.default_speed, network.default_capacity)
                link_dict[link.link_id] = link
                max_link_id += 1
                if not way.oneway:
                    linkb = Link(max_link_id)
                    linkb.buildFromOSMWay(way, -1, list(reversed(m_segment_node_list)), network.default_lanes, network.default_speed, network.default_capacity)
                    link_dict[linkb.link_id] = linkb
                    max_link_id += 1
    network.link_dict = link_dict
    network.max_link_id = max_link_id


def _identifyPureCycleWays(link_way_list):
    for way in link_way_list:
        if way.is_cycle:
            way.is_pure_cycle = True
            for node in way.ref_node_list[1:-1]:
                if node.is_crossing:
                    way.is_pure_cycle = False
                    break


def _addSignalFromLink(network):
    for link_id, link in network.link_dict.items():
        if link.ctrl_type == 'signal':
            to_node = link.to_node
            if to_node.ctrl_type != 'signal':
                if len(to_node.incoming_link_list) > 1 or len(to_node.outgoing_link_list) > 1:
                    to_node.ctrl_type = 'signal'


def _removeIsolated(network, min_nodes):
    if og_settings.verbose:
        print(f'    removing sub networks with less than {min_nodes} nodes (mrnet)')

    node_list = []
    node_to_idx_dict = {}
    for idx, (node_id,node) in enumerate(network.node_dict.items()):
        node_list.append(node)
        node_to_idx_dict[node] = idx

    number_of_nodes = len(node_list)
    node_group_id_list = [-1] * number_of_nodes

    group_id = 0
    start_idx = 0

    while True:
        unprocessed_node_list = [node_list[start_idx]]
        node_group_id_list[start_idx] = group_id
        while unprocessed_node_list:
            node = unprocessed_node_list.pop()
            for ob_link in node.outgoing_link_list:
                ob_node = ob_link.to_node
                if node_group_id_list[node_to_idx_dict[ob_node]] == -1:
                    node_group_id_list[node_to_idx_dict[ob_node]] = group_id
                    unprocessed_node_list.append(ob_node)

            for ib_link in node.incoming_link_list:
                ib_node = ib_link.from_node
                if node_group_id_list[node_to_idx_dict[ib_node]] == -1:
                    node_group_id_list[node_to_idx_dict[ib_node]] = group_id
                    unprocessed_node_list.append(ib_node)

        unreachable_node_exits = False
        idx = 0
        for idx in range(start_idx+1,number_of_nodes):
            if node_group_id_list[idx] == -1:
                unreachable_node_exits = True
                break

        if unreachable_node_exits:
            start_idx = idx
            group_id += 1
        else:
            break

    group_id_set = set(node_group_id_list)
    group_isolated_dict = {}
    for group_id in group_id_set:
        group_size = node_group_id_list.count(group_id)
        if group_size < min_nodes:
            group_isolated_dict[group_id] = True
        else:
            group_isolated_dict[group_id] = False

    removal_link_set = set()
    for idx, node in enumerate(node_list):
        if group_isolated_dict[node_group_id_list[idx]]:
            del network.node_dict[node.node_id]
            for ob_link in node.outgoing_link_list: removal_link_set.add(ob_link)
            for ib_link in node.incoming_link_list: removal_link_set.add(ib_link)
    for link in removal_link_set:
        del network.link_dict[link.link_id]


def _preprocessWays(osmnetwork, link_types, network_types):
    link_way_list = []
    POI_way_list = []
    network_types_set = set(network_types)
    include_railway = True if 'railway' in network_types_set else False
    include_aeroway = True if 'aeroway' in network_types_set else False

    for _, way in osmnetwork.osm_way_dict.items():
        if way.building or way.amenity or way.leisure:
            POI_way_list.append(way)
        elif way.highway:
            if way.highway in highway_poi_set:
                way.way_poi = way.highway
                POI_way_list.append(way)
                continue
            if way.area and way.area != 'no':
                continue
            if way.highway in negligible_highway_type_set:
                continue
            if len(way.ref_node_list) < 2:
                continue
            try:
                way.link_type_name, way.is_link = og_settings.osm_highway_type_dict[way.highway]
                way.link_type = og_settings.link_type_no_dict[way.link_type_name]
            except KeyError:
                logger = getLogger()
                if logger: logger.warning(f'new highway type at way {way.osm_way_id}, {way.highway}')
                continue
            valid_link_type = link_types == 'all' or way.link_type_name in link_types
            if not valid_link_type:
                continue
            allowable_agent_type_list = getAllowableAgentType(way)
            way.allowable_agent_type_list = allowable_agent_type_list
            if len(way.allowable_agent_type_list) == 0:
                continue
            way.allowed_uses = way.allowable_agent_type_list
            way.ref_node_list[0].is_crossing = True
            way.ref_node_list[-1].is_crossing = True
            for node in way.ref_node_list:
                node.usage_count += 1
            if way.ref_node_list[0] is way.ref_node_list[-1]:
                way.is_cycle = True
            if way.oneway is None:
                if way.junction in ['circular', 'roundabout']:
                    way.oneway = True
                else:
                    way.oneway = og_settings.default_oneway_flag_dict[way.link_type_name]
            way.link_class = 'highway'
            link_way_list.append(way)
        elif way.railway:
            if not include_railway: continue
            if way.railway in negligible_railway_type_set:
                continue
            if len(way.ref_node_list) < 2:
                continue
            way.ref_node_list[0].is_crossing = True
            way.ref_node_list[-1].is_crossing = True
            for node in way.ref_node_list[1:-1]: node.usage_count += 1
            way.link_type_name = way.railway
            way.link_type = og_settings.link_type_no_dict['railway']
            if way.oneway is None:
                way.oneway = og_settings.default_oneway_flag_dict['railway']
            way.link_class = 'railway'
            link_way_list.append(way)
        elif way.aeroway:
            if not include_aeroway: continue
            if way.aeroway in negligible_aeroway_type_set:
                continue
            if len(way.ref_node_list) < 2:
                continue
            way.ref_node_list[0].is_crossing = True
            way.ref_node_list[-1].is_crossing = True
            for node in way.ref_node_list[1:-1]: node.usage_count += 1
            way.link_type_name = way.aeroway
            way.link_type = og_settings.link_type_no_dict['aeroway']
            if way.oneway is None:
                way.oneway = og_settings.default_oneway_flag_dict['aeroway']
            way.link_class = 'aeroway'
            link_way_list.append(way)
        else:
            pass

    osmnetwork.link_way_list = link_way_list
    osmnetwork.POI_way_list = POI_way_list


def _identifyCrossingOSMNodes(osm_node_dict):
    for _, osmnode in osm_node_dict.items():
        if osmnode.usage_count >= 2 or osmnode.ctrl_type == 'signal':
            osmnode.is_crossing = True


def _buildNet(osmnetwork, network_types, link_types, POI, POI_percentage, offset, min_nodes, combine, bbox,
              default_lanes, default_speed, default_capacity, start_node_id, start_link_id):
    if og_settings.verbose:
        print('  parsing osm network (mrnet)')

    network = Network()
    network.max_node_id = start_node_id
    network.max_link_id = start_link_id

    network.GT, network.bounds = osmnetwork.GT, osmnetwork.bounds

    if isinstance(default_lanes, dict):
        network.default_lanes = default_lanes
    elif default_lanes is True:
        network.default_lanes = og_settings.default_lanes_dict
    if isinstance(default_speed, dict):
        network.default_speed = default_speed
    elif default_speed is True:
        network.default_speed = og_settings.default_speed_dict
    if isinstance(default_capacity, dict):
        network.default_capacity = default_capacity
    elif default_capacity is True:
        network.default_capacity = og_settings.default_capacity_dict

    _preprocessWays(osmnetwork, link_types, network_types)
    _identifyCrossingOSMNodes(osmnetwork.osm_node_dict)
    _identifyPureCycleWays(osmnetwork.link_way_list)
    _createNodesAndLinks(network, osmnetwork.link_way_list)
    _addSignalFromLink(network)
    if POI:
        from .pois import generatePOIs
        generatePOIs(osmnetwork.POI_way_list, osmnetwork.osm_relation_list, network, POI_percentage)

    if min_nodes > 1: _removeIsolated(network, min_nodes)
    if combine: combineShortLinks(network)

    if offset != 'no':
        if og_settings.verbose:
            print('    offseting link geometries (mrnet)')
        distance_sp = -2 if offset == 'right' else 2
        distance_ma = 2 if offset == 'right' else -2
        GT = network.GT
        for _, link in network.link_dict.items():
            if getattr(link, 'from_bidirectional_way', False):
                geometry_xy = link.geometry_xy.offset_curve(distance=distance_sp, join_style=2)
                if isinstance(geometry_xy, geometry.MultiLineString):
                    geometry_xy = offsetLine(link.geometry_xy, distance_ma)
                link.geometry_xy = geometry_xy
                link.geometry = GT.geo_to_latlon(link.geometry_xy)

    return network


def getNetFromFile(filename='map.osm', network_types=('auto',), link_types='all', POI=False, POI_sampling_ratio=1.0,
                   strict_mode=True, offset='no', min_nodes=1, combine=False, bbox=None,
                   default_lanes=False, default_speed=False, default_capacity=False, start_node_id=0, start_link_id=0):
    if og_settings.verbose:
        print('arguments used for network parsing (mrnet):')
        print(f'  filename: {filename}')
        print(f'  network_types: {network_types}')
        print(f'  link_types: {link_types}')
        print(f'  POI: {POI}')
        print(f'  POI_sampling_ratio: {POI_sampling_ratio}')
        print(f'  strict_mode: {strict_mode}')
        print(f'  offset: {offset}')
        print(f'  min_nodes: {min_nodes}')
        print(f'  combine: {combine}')
        print(f'  bbox: {bbox}')
        print(f'  default_lanes: {default_lanes}')
        print(f'  default_speed: {default_speed}')
        print(f'  default_capacity: {default_capacity}')
        print(f'  start_node_id: {start_node_id}')
        print(f'  start_link_id: {start_link_id}\n')

        print('Building Network from OSM file (mrnet)')

    network_types_, link_types_, POI_, POI_sampling_ratio_, strict_mode_, offset_, min_nodes_, combine_, \
        bbox_, default_lanes_, default_speed_, default_capacity_, start_node_id_, start_link_id_ = \
        checkArgs_getNetFromFile(filename, network_types, link_types, POI, POI_sampling_ratio, strict_mode, offset,
                                min_nodes, combine, bbox, default_lanes, default_speed, default_capacity, start_node_id,
                                 start_link_id)

    osmnetwork = readOSMFile(filename, POI_, strict_mode_, bbox_)
    network = _buildNet(osmnetwork, network_types_, link_types_, POI_, POI_sampling_ratio_, offset_, min_nodes_, combine_,
                        bbox_, default_lanes_, default_speed_, default_capacity_, start_node_id_, start_link_id_)

    if og_settings.verbose:
        print(f'  number of nodes: {len(network.node_dict)}, number of links: {len(network.link_dict)}, number of pois: {len(network.POI_list)}')

    return network


